/*
 * Copyright 2011 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <omp.h>

#define STENCIL_SIZE (16 * 1024 * 1024)
static const int RADIUS = 15;

#define ELEMENTS(S)  ((S)-2*RADIUS)

#define BLOCK_SIZE 512
#define NUM_BLOCKS(S) (((S)+BLOCK_SIZE-1)/BLOCK_SIZE)

#define NUM_CHUNKS 16
#define NUM_STREAMS 4

double inline CLOCK() {
  return omp_get_wtime();
}

void randomizeArray(float *in, int N) {
  for(int i=0;i<N;i++)
    in[i]=rand()/float(RAND_MAX);
}

#define e 1e-4
void checkAnswers(char *label, int N, float *gold, float *data) {
  bool pass=true;
  for (int i=RADIUS;i<N-RADIUS;i++)
    if (fabs(gold[i]-data[i])>e) {
     pass=false; 
     printf("%s, Fail at idx: %d, gold: %f, val: %f\n",label, i,gold[i],data[i]);
     return;
  }
  if (!pass) printf("FAIL\n");
}

void applyStencil1D_cpu(int startIdx, int endIdx, const float *weights,
                        float *in, float *out) {
  for(int i=startIdx;i<endIdx;i++) {  //for each element
    float result=0;
    for (int j=-RADIUS;j<=RADIUS;j++) {
      result += weights[j+RADIUS] * in[i+j];
    }
    out[i] = result / (2 * RADIUS + 1);
  }
}

#ifndef GPU
void applyStencil1D_cpu_opt(int startIdx, int endIdx, const float *__restrict__ weights,
                            float * __restrict__ in, float * __restrict__ out) {
  #pragma omp parallel for
  #pragma prefetch in
  #pragma vector always nontemporal
  for(int i=startIdx;i<endIdx;i++) {  //for each element
    float result=0;
    #pragma unroll
    for (int j=-RADIUS;j<=RADIUS;j++) {
      result += weights[j+RADIUS] * in[i+j];
    }
    out[i] = result / (2 * RADIUS + 1);
  }
}

static void
cpu(size_t N, const float *weights, float *in, float *gold)
{
  int size = N * sizeof(float);
  float *out = (float*)malloc(size);

  double time_start = CLOCK();
  applyStencil1D_cpu(RADIUS, N-RADIUS, weights, in, out);
  double elapsed_time = CLOCK() - time_start;
  checkAnswers("CPU", N, gold, out);
  printf("    CPU: %lg seconds, %lg GBytes/s, %lg GElements/s\n", 
         elapsed_time, 
         N * sizeof(float) * 2 / elapsed_time / 1e9,
         ELEMENTS(N) / elapsed_time / 1e9);

  free(out);
}

static void
cpu_opt(size_t N, const float *weights, float *in, float *gold)
{
  int size = N * sizeof(float);
  float *out = (float*)malloc(size);

  double time_start = CLOCK();
  applyStencil1D_cpu_opt(RADIUS, N-RADIUS, weights, in, out);
  double elapsed_time = CLOCK() - time_start;
  checkAnswers("CPU OPT", N, gold, out);
  printf("    CPU OPT: %lg seconds, %lg GBytes/s, %lg GElements/s\n", 
         elapsed_time, 
         N * sizeof(float) * 2 / elapsed_time / 1e9,
         ELEMENTS(N) / elapsed_time / 1e9);

  free(out);
}
#endif

#ifdef GPU
__global__ void applyStencil1D(int startIdx, int endIdx, const float *weights, 
                               float *in, float *out) {
  int i = startIdx + blockIdx.x*blockDim.x+threadIdx.x;
 
  if (i < endIdx) {
    float result = 0;
    for (int j= -RADIUS; j <= RADIUS; j++) {
      result += weights[j+RADIUS] * in[i+j];
    }
    out[i] = result / (2 * RADIUS + 1);
  }
}

static void
gpu(size_t N, const float *weights, float *in, float *gold)
{
  int size = N * sizeof(float);
  int weights_size = (RADIUS * 2  + 1) * sizeof(float);
  float *d_weights, *d_in, *d_out;
  float *out = (float*)malloc(size);
  cudaMalloc(&d_weights, weights_size);
  cudaMalloc(&d_in, size);
  cudaMalloc(&d_out, size);

  cudaDeviceSynchronize();
  double time_start = CLOCK();

  cudaMemcpy(d_weights, weights, weights_size, cudaMemcpyHostToDevice);
  cudaMemcpy(d_in, in, size, cudaMemcpyHostToDevice);
  applyStencil1D<<<NUM_BLOCKS(N),BLOCK_SIZE>>>(RADIUS, N-RADIUS, d_weights, d_in, d_out);
  cudaMemcpy(out, d_out, size, cudaMemcpyDeviceToHost); 

  cudaDeviceSynchronize();
  double elapsed_time = CLOCK() - time_start;
  checkAnswers("GPU", N, gold, out);
  printf("    GPU: %lg seconds, %lg GBytes/s, %lg GElements/s\n", 
         elapsed_time, 
         N * sizeof(float) * 2 / elapsed_time / 1e9,
         ELEMENTS(N) / elapsed_time / 1e9);

  cudaFree(d_weights);
  cudaFree(d_in);
  cudaFree(d_out);
  free(out);
}

static void
gpu_pinned(size_t N, const float *weights, float *in, float *gold)
{
  int size = N * sizeof(float);
  int weights_size = (RADIUS * 2  + 1) * sizeof(float);
  float *d_weights, *d_in, *d_out;
  float *h_in, *h_out;
  cudaMallocHost(&h_in, size);
  cudaMallocHost(&h_out, size);
  cudaMalloc(&d_weights, weights_size);
  cudaMalloc(&d_in, size);
  cudaMalloc(&d_out, size);

  memcpy(h_in, in, size);

  cudaDeviceSynchronize();
  double time_start = CLOCK();

  cudaMemcpy(d_weights, weights, weights_size, cudaMemcpyHostToDevice);
  cudaMemcpy(d_in, h_in, size, cudaMemcpyHostToDevice);
  applyStencil1D<<<NUM_BLOCKS(N),BLOCK_SIZE>>>(RADIUS, N-RADIUS, d_weights, d_in, d_out);
  cudaMemcpy(h_out, d_out, size, cudaMemcpyDeviceToHost); 

  cudaDeviceSynchronize();
  double elapsed_time = CLOCK() - time_start;
  checkAnswers("GPU PINNED", N, gold, h_out);
  printf("    GPU PINNED: %lg seconds, %lg GBytes/s, %lg GElements/s\n", 
         elapsed_time, 
         N * sizeof(float) * 2 / elapsed_time / 1e9,
         ELEMENTS(N) / elapsed_time / 1e9);

  cudaFree(d_weights);
  cudaFree(d_in);
  cudaFree(d_out);
  cudaFreeHost(h_in);
  cudaFreeHost(h_out);
}

static void
gpu_overlap(size_t N, const float *weights, float *in, float *gold)
{
  int size = N * sizeof(float);
  int weights_size = (RADIUS * 2  + 1) * sizeof(float);
  int elementsPerChunk = N / NUM_CHUNKS;
  float *d_weights, *d_in, *d_out;
  float *h_in, *h_out;
  cudaMallocHost(&h_in, size);
  cudaMallocHost(&h_out, size);
  cudaMalloc(&d_weights, weights_size);
  cudaMalloc(&d_in, size);
  cudaMalloc(&d_out, size);

  memcpy(h_in, in, size);

  cudaStream_t streams[NUM_STREAMS];
  for (int s = 0; s < NUM_STREAMS; s++)
    cudaStreamCreate(streams + s);

  cudaDeviceSynchronize();
  double time_start = CLOCK();
  cudaMemcpy(d_weights, weights, weights_size, cudaMemcpyHostToDevice);

  int copyStart = 0, copyEnd = elementsPerChunk;
  int computeStart = RADIUS, computeEnd = elementsPerChunk-RADIUS;
  for (int i = 0; i < NUM_CHUNKS; i++) {
    cudaMemcpyAsync(d_in+copyStart, h_in+copyStart, (copyEnd-copyStart)*sizeof(float), 
                    cudaMemcpyHostToDevice, streams[i % NUM_STREAMS]);
    applyStencil1D<<<NUM_BLOCKS(computeEnd-computeStart), 
      BLOCK_SIZE, 0, streams[i % NUM_STREAMS]>>>(computeStart, computeEnd, d_weights, d_in, d_out);
    cudaMemcpyAsync(h_out+computeStart, d_out+computeStart, (computeEnd-computeStart)*sizeof(float),
                    cudaMemcpyDeviceToHost, streams[i % NUM_STREAMS]); 
    copyStart = copyEnd;
    copyEnd = (i == (NUM_CHUNKS-1)) ? N : copyStart + elementsPerChunk;
    computeStart = computeEnd;
    computeEnd = copyEnd - RADIUS;
  }
  cudaDeviceSynchronize();
  double elapsed_time = CLOCK() - time_start;
  checkAnswers("GPU OVERLAP", N, gold, h_out);
  printf("    GPU OVERLAP: %lg seconds, %lg GBytes/s, %lg GElements/s\n", 
         elapsed_time, 
         N * sizeof(float) * 2 / elapsed_time / 1e9,
         ELEMENTS(N) / elapsed_time / 1e9);

  cudaFree(d_weights);
  cudaFree(d_in);
  cudaFree(d_out);
  cudaFreeHost(h_in);
  cudaFreeHost(h_out);
}
#endif

int 
main()
{
  int size = STENCIL_SIZE * sizeof(float);
  float *weights = (float *)malloc((RADIUS * 2 + 1) * sizeof(float));
  float *in = (float *)malloc(size);
  float *gold = (float *)malloc(size);
  
  // calculate expected output
  randomizeArray(weights, RADIUS * 2 + 1);
  randomizeArray(in, STENCIL_SIZE);
  applyStencil1D_cpu(RADIUS, STENCIL_SIZE-RADIUS, weights, in, gold); 
  
#ifndef GPU
  // run serial cpu
  cpu(STENCIL_SIZE, weights, in, gold);

  // run parallelized cpu
  cpu_opt(STENCIL_SIZE, weights, in, gold);

#else
  // run gpu
  gpu(STENCIL_SIZE, weights, in, gold);

  // run gpu w/ pinned memory
  gpu_pinned(STENCIL_SIZE, weights, in, gold);

  // run gpu w/ overlapped memory and compute
  gpu_overlap(STENCIL_SIZE, weights, in, gold);

#endif

  free(weights);
  free(in);
  free(gold);
}
